"""
Implements control laws formed by control lyapunov functions (CLF)
for processes that are affine in the control:

x_dot = f(x) + g(x)u

(Let x \in \reals^m denote state, u \in \reals^n denote control,
so g(x) \in \reals^{m,n}).
The CLFs are based on feedback linearization, which looks at the
output we seek to control, say,

drive y(x) = Cx - d -> 0

and chooses the control u based on the the output's derivative

y_dot = L_f y(x) + L_g y(x) u

where L_f is the Lie derivative wrt f(x); L_f y(x) = dy(x)/dx f(x).
The control is chosen like u = [L_g y(x)]^-1 (-L_f y(x) + mu) to obtain

y_dot = mu

This mu can then be chosen to satisfy desirable CLF properties such
as exponential stabilization.
"""
import numpy as np
from misc import riccati
from scipy import optimize as opt


class MinNormRd1ClfController(object):
    """
    Let x \in \reals^m denote state, y \in \reals^r be an output we
    seek to control.
    Use a CLF to drive
    y = Cx - d -> 0
    at a given exponential rate, selecting the minimum norm control.
    This controller applies only for y with a relative degree 1,
    i.e., if u = acceleration, y must be velocities.

    Specifically, the set of controls satisfying the rate is taken as:
    {u \in \reals^n | L_f V(x) + L_g V(x) u + eps V(x) <= 0}
    =
    {u \in \reals^n | phi_0(x) + phi_1(x)' u <= 0}
    where
    V(x) = y(x)'y(x)
    phi_0(x) = L_f V(x) + eps V(x)
             = 2y'Cf(x) + eps y'y
    phi_1(x) = (L_g V(x) )'
             = [2y'Cg(x)]'
    The minimum norm control is given by choosing the minimum norm element
    of the set. The minimum norm control is
    u = {  0                                  if phi_0(x) <= 0
         -phi_0(x)phi_1(x)/||phi_1(x)||^2_2   else

    This implementation closely follows material from:
    A. D. Ames, J. W. Grizzle, and P. Tabuada. "Control barrier function based
    quadratic programs with application to adaptive cruise control," in
    IEEE Conference on Decision and Control (CDC), 2014, pp. 6271-6278.
    """

    def __init__(self, eps_rate):
        """
        :param eps_rate: rate for exponential convergence to setpoint
        """
        self.eps_rate = eps_rate
        self.C = np.zeros(())
        self.d = np.zeros(())

    def set_setpoint(self, C, d):
        self.C = C
        self.d = d

    def __call__(self, x, t, fx=None, gx=None, **kwargs):
        y = self.C.dot(x) - self.d
        phi_zero = 2 * y.T.dot(self.C).dot(fx) + self.eps_rate * y.dot(y)  # scalar
        phi_one = 2 * y.T.dot(self.C).dot(gx)  # n,
        if phi_zero <= 0:
            return 0 * phi_one
        bot = np.linalg.norm(phi_one) ** 2
        u = -phi_zero * phi_one / max(bot, 1e-8)
        return u


class MinNormRd2ClfController(object):
    """
    Let x \in \reals^m denote state, y \in \reals^r be an output we
    seek to control.
    Use a CLF to drive
    y = Cx - d -> 0
    at a given exponential rate, selecting the minimum norm control.
    This controller applies only for y with a relative degree 2,
    i.e., if u = acceleration, y must be position/heading.

    Compared to the case of relative degree 1, we need to take one
    more derivative to get to the form y_dot = mu.

    This implementation closely follows material from:
    A. D. Ames and M. Powell. "Towards the unification of locomotion and
    manipulation through control lyapunov functions and quadratic programs,"
    in Control of Cyber-Physical Systems. Springer, 2013, pp. 219-240.
    """

    def __init__(self, eps_rate, c_delta, u_min, u_max):
        """
        :param eps_rate: rate for exponential convergence to setpoint
        :param c_delta: penalty for violating CLF constraint
        :param u_min: n, | minimum input values
        :param u_max: n, | maximum input values
        """
        self.eps_rate = eps_rate
        self.c_delta = c_delta
        self.u_min = u_min
        self.u_max = u_max
        self.C = np.zeros(())
        self.d = np.zeros(())
        self.C_dot = np.zeros(())  # for extracting y_dot(x_t) from x_t
        self.F = np.zeros(())
        self.G = np.zeros(())
        self.Q = np.zeros(())
        self.P = np.zeros(())

    def set_setpoint(self, C, d, C_dot, Q=()):
        self.C = C
        self.d = d
        self.C_dot = C_dot
        r = C.shape[0]
        self.F = np.zeros((2*r, 2*r))
        self.F[:r, r:] = np.eye(r)
        self.G = np.zeros((2*r, r))
        self.G[r:, :r] = np.eye(r)
        self.Q = np.eye(2*r) if len(Q) == 0 else Q  # equal weighting
        self.P = riccati.solve_cts_riccati_schur(
            self.F, self.G, self.Q, np.eye(r))  # 2r, 2r

    def __call__(self, x, t, fx=None, gx=None, dfxdx=None,
                 A_cbf=(), b_cbf=(), **kwargs):
        """
        :param x:
        :param t:
        :param fx: m, |
        :param gx: m, n |
        :param dfxdx: m, m | jacobian of fx
        :param A_cbf: k, n | for k control barrier function
            constraints, if specified
        :param b_cbf: k,
        :param kwargs:
        :return:
        """
        # CLF
        y = self.C.dot(x) - self.d
        y_dot = self.C_dot.dot(x)
        eta = np.hstack((y, y_dot))
        V = eta.dot(self.P).dot(eta)
        PF = self.P.dot(self.F)
        LfV = eta.dot((PF.T + PF).dot(eta))
        LgV = 2*eta.dot(self.P).dot(self.G)
        phi_zero = LfV + self.eps_rate * V
        phi_one = LgV
        # input constraints
        dLfydx = self.C.dot(dfxdx)  # r, m
        A = dLfydx.dot(gx)  # r, n
        # at minimum need r <= m,n for right inverse, so u is defined
        # - should have this inverse by relative degree 2 assumption
        Ainv = np.linalg.pinv(A)  # n, r
        Lf = dLfydx.dot(fx)  # r,
        AinvLf = Ainv.dot(Lf)  # n,
        # solve the QP, writing constraints as Ac z <= bc
        r = y.size
        z0 = np.zeros((r+1,))  # z = [mu, delta]
        hess_vec = np.ones((r+1,))
        hess_vec[-1] = self.c_delta
        Ac = np.vstack((phi_one, Ainv, -Ainv))
        Ac = np.hstack((Ac, 0 * Ac[:, [0]]))  # for delta
        Ac[0, -1] = -1
        bc = np.hstack((-phi_zero, self.u_max + AinvLf, -self.u_min - AinvLf))
        if len(A_cbf) > 0:
            A_cbf_mu = A_cbf.dot(Ainv)  # k, r
            A_cbf_mu = np.hstack((A_cbf_mu, 0 * A_cbf_mu[:, [0]]))
            Ac = np.vstack((Ac, A_cbf_mu))
            bc = np.hstack((bc, b_cbf + A_cbf.dot(AinvLf)))
        linear_constraint = opt.LinearConstraint(Ac, bc - np.inf, bc)
        res = opt.minimize(
            lambda z: 0.5 * (self.c_delta * z[-1]**2 + np.linalg.norm(z[:-1])**2),
            z0, method='trust-constr',
            jac=lambda z, *args: hess_vec * z,
            hessp=lambda x, p, *args: hess_vec * p,
            constraints=[linear_constraint, ],
        )
        z_star = res.x
        mu_star = z_star[:-1]
        u_star = Ainv.dot(-Lf + mu_star)
        return u_star


class MinNormRd2CbfObstacleController(object):
    """
    Use a control barrier function (CBF) to enforce a minimum distance
    to point obstacles.
    Assumption:
        The control input affects acceleration so the CBF constraint on
        position is of relative degree two.

    To avoid obstacles with distance d, use barrier function:
    b_i(x_t) = ||p_t - p_i||^2_2 - d^2
    where p_t = Cx_t \in \reals^2 is current position
    and p_i \in \reals^2 is the ith obstacle's position.
    The full barrier function for the k obstacles is
    b(x_t) = (b_1(x_t); ...; b_k(x_t))

    The CBF is enforced using linear class K functions having
    the same coefficient. This controller basically makes the
    constraints corresponding to the CBF, and hands these to
    a CLF-based controller to use when selecting controls.

    This implementation closely follows material from:
    W. Xiao and C. Belta. "Control barrier functions for systems with
    high relative degree," in IEEE Conference on Decision and Control
    (CDC), 2019, pp. 474-479.
    """
    def __init__(self, rd2clf_controller, C, c_alpha, d):
        """
        :param rd2clf_controller: clf controller for systems of relative degree 2
        :param C: 2, m | selects position from current state, p_t = Cx_t
        :param c_alpha: coefficient for linear class K functions
        :param d: minimum distance allowed to each obstacle
        """
        self.rd2clf_controller = rd2clf_controller
        self.C = C
        self.c_alpha = c_alpha
        self.d = d
        self.p = np.zeros((0, 2))

    def set_obstacles(self, p):
        """
        :param p: k, 2 | obstacle positions
        :return:
        """
        self.p = p.reshape(-1, 2)

    def set_setpoint(self, *args, **kwargs):
        self.rd2clf_controller.set_setpoint(*args, **kwargs)

    def __call__(self, x, t, fx=None, gx=None, dfxdx=None, **kwargs):
        m = x.size
        d_t = self.C.dot(x) - self.p  # k, 2
        b = np.linalg.norm(d_t, axis=1)**2 - self.d**2  # k,
        dbdx = 2 * d_t.dot(self.C)  # k, m
        Lfb = dbdx.dot(fx)  # k,

        # only enforce barrier for valid initial conditions
        is_phi_zero = b >= 0
        is_phi_one = Lfb + self.c_alpha * b >= 0
        is_valid = is_phi_zero & is_phi_one
        k = is_valid.sum()
        if k < 1:
            return self.rd2clf_controller(x, t, fx=fx, gx=gx, dfxdx=dfxdx, **kwargs)

        b = b[is_valid]
        dbdx = dbdx[is_valid, :]
        Lfb = Lfb[is_valid]
        d2bdx2_i = 2 * self.C.T.dot(self.C)  # m, m since d^2b/dx^2 is (k, m, m) tensor
        dLfbdx = np.broadcast_to(d2bdx2_i.dot(fx), (k, m)) \
            + dbdx.dot(dfxdx)  # k, m
        Lf2b = dLfbdx.dot(fx)  # k,
        LgLfb = dLfbdx.dot(gx)  # k, n
        b_cbf = Lf2b + 2*self.c_alpha*Lfb + self.c_alpha**2 * b
        A_cbf = -LgLfb
        return self.rd2clf_controller(x, t, fx=fx, gx=gx, dfxdx=dfxdx,
                                      A_cbf=A_cbf, b_cbf=b_cbf, **kwargs)


def make_C_matrix(select_inds, m):
    """
    Build C matrix for common choices of y = Cx - d where
    y is a selection of x's values (such as position) and
    d = y_desired (so Cx - d = 0 implies we are at the desired position)
    :param select_inds: list of indices of state x
    :param m: dimension of state
    :return: r, n | C matrix
    """
    r = len(select_inds)
    C = np.zeros((r, m))
    C[np.arange(r), select_inds] = 1.
    return C
