"""
Implements state-dependent riccati equation (SDRE) approach
for continuous-time infinite-horizon control.

Let x \in \reals^n denote the current state and x_0 denote the
target state (the time subscript for x is dropped).
Let x_e = x - x0.
The SDRE approach finds a sequence of controls u \in \reals^m
(time subscript dropped here too) to minimize the 'cost' functional

J(x, u) = \integral_{t_0}^\infty  x'Qx + u'Ru dt,

where system state evolves according to the nonlinear differential constraint

x_dot = A(x)x + B(x)u.

For an overview of SDRE methods, see
J. R. Cloutier, "State-dependent riccati equation techniques: an overview,"
in Proceedings of the American Control Conference, 1997, pp. 932-936.
"""
import numpy as np
from misc import riccati


class SdreController(object):

    def __init__(self, Q, R):
        """
        :param Q:
        :param R:
        """
        self.Q = Q
        self.R = R
        self.x_goal = ()

    def set_setpoint(self, x_goal):
        self.x_goal = x_goal

    def __call__(self, x, t, A=(), B=(), **kwargs):
        Rinv = np.linalg.inv(self.R)
        P = riccati.solve_cts_riccati_schur(A, B, self.Q, Rinv)
        u = -Rinv.dot(B.T.dot(P.dot(x - self.x_goal)))
        return u

