import numpy as np
from scene_elements.scene import Scene
from scene_elements.sphere import Sphere
from scene_elements.water import Water
from scene_elements.camera import Camera
from misc.interpolation import build_nd_interpolator
from scipy.interpolate import InterpolatedUnivariateSpline
from rigid_body_models import boat
from vis import planar_thruster_model
import misc.matrix_building as mb
import matplotlib.pyplot as plt


def xyplane_projection_error_cost_surface(P_array, uv_array, xyz_true, z_level=0):
    """
    Projection error model:

    (u_i; v_i; 1) = P_i (x; y; z; 1) + e_i   [with e_i = (-, -, 0)]

    Plot projection error (sum of squared e_i) surface by calculating error at various xyz.
    Grid xyz by centering xy on true xy, and using z_level for all z.
    :param P_array: k, 3, 4
    :param uv_array: k, 2
    :param xyz_true: 3,
    :param z_level:
    :return:
        x_grid: n, m | x coordinates of tested xyz
        y_grid: n, m | y coordinates of tested xyz
        sse_grid: n, m | sum of squared errors at each tested xyz
    """
    xyz1_test = np.mgrid[-10:10:.1, -10:10:.1]
    xyz1_test[0] += xyz_true[0]
    xyz1_test[1] += xyz_true[1]
    m, n = xyz1_test.shape[1:]
    xyz1 = np.concatenate((
        xyz1_test, z_level * xyz1_test[[0]], 1 + 0 * xyz1_test[[0]]))
    xyz1 = xyz1.reshape(4, -1)  # 4, m*n
    sse = np.zeros((m*n))
    for i in range(uv_array.shape[0]):
        uv_proj = P_array[i].dot(xyz1).T  # m*n, 3
        np.clip(uv_proj[:, 2], 1e-5, None, out=uv_proj[:, 2])  # avoid / 0
        uv_proj = uv_proj[:, :2] / uv_proj[:, [2]]
        sse += np.linalg.norm(uv_proj - uv_array[i], axis=1) ** 2
    return xyz1_test[0], xyz1_test[1], sse.reshape(m, n)


def main():
    target_sphere = Sphere([2, 6], color='yellow')

    scene = Scene([
        Sphere([1, 2], color='green'),
        Sphere([-1, 2], color='red'),
        target_sphere,
        Sphere([0, 7], color='black'),
        Water(),
    ])

    cam_f = 420.
    cam_rows = 480
    cam_cols = 640
    cam_K = np.array([
        [cam_f, 0, cam_rows/2, 0],
        [0, cam_f, cam_cols/2, 0],
        [0, 0, 1., 0],
    ])
    cam_Rt = np.array([
        [0, 0, 1, -.2],
        [0, -1, 0, 0],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
    ])
    camera = Camera(K=cam_K, Rt=cam_Rt, cam_cols=cam_cols, cam_rows=cam_rows)

    DT = 0.5
    cam_traj_xytheta_t_points = np.array([
        [0, 0, np.pi / 2, 0],
        [.5, 5, np.pi / (2 + .5), 10],
    ])
    cam_traj_spline = build_nd_interpolator(
        cam_traj_xytheta_t_points[:, :-1], cam_traj_xytheta_t_points[:, -1],
        InterpolatedUnivariateSpline, k=1,
    )
    xytheta = cam_traj_spline(np.mgrid[0:cam_traj_xytheta_t_points[-1, -1]:DT])

    # For displaying the trajectory
    # u = np.empty((xytheta.shape[0], 0))
    # boat_model = boat.DiamondMountedThrusters()
    # vis_mounted_thruster_model.loop_camera_sim(
    #     boat_model, xytheta[:, [0, 1]], xytheta[:, 2], u, DT, scene, camera)
    # -

    n_steps = xytheta.shape[0]
    P_array = np.zeros((n_steps, 3, 4))
    uv_array = np.zeros((n_steps, 2))
    for i in range(n_steps):
        body2world_Rt = mb.Rt_planar_xyz1(xytheta[i, -1], xytheta[i, :2])
        camera.set_body2world_Rt(body2world_Rt)
        P_array[i] = camera.get_world_projection_matrix()
        uv_array[i] = target_sphere.calculate_image_view_center(camera)

    # Select indices where target is in image
    mask = ~np.isnan(uv_array).any(axis=1)
    P_array, uv_array = P_array[mask], uv_array[mask]
    mask = (0 <= uv_array[:, 0]) & (uv_array[:, 0] < cam_rows) & \
           (0 <= uv_array[:, 1]) & (uv_array[:, 1] < cam_cols)
    P_array, uv_array = P_array[mask], uv_array[mask]

    n_obs = uv_array.shape[0]
    print('{} observations of target collected'.format(n_obs))

    xyz_true = np.hstack([target_sphere.xy, target_sphere.z])
    x_grid, y_grid, sse_grid = xyplane_projection_error_cost_surface(
        P_array, uv_array, xyz_true, z_level=0)
    sse_grid /= 1e6  # scale by a fictitious variance
    sse_grid[sse_grid > 1e2] = 1e2

    # 2d display
    fig, ax = plt.subplots()
    ax.imshow(sse_grid, extent=(x_grid.min(), x_grid.max(), y_grid.min(), y_grid.max()),
              origin='lower', vmin=0, vmax=1)
    ax.set_xlabel('x [m]')
    ax.set_ylabel('y [m]')
    ax.scatter(*target_sphere.xy, marker='+', color='w')
    plt.show()

    # 3d display
    # from mpl_toolkits.mplot3d import Axes3D
    # fig = plt.figure()
    # ax = fig.add_subplot(111, projection='3d')
    # surf = ax.plot_surface(x_grid, y_grid, sse_grid, cmap='jet',
    #                        linewidth=0, antialiased=False, vmin=0, vmax=50)
    # ax.set_xlabel('x [m]')
    # ax.set_ylabel('y [m]')
    # plt.show()

    # Estimate x
    import examples.monocular_mapper as mm

    x_hat = mm.dlt_estimate(P_array, uv_array)
    print('Estimated position: {}'.format(x_hat.round(2)))


if __name__ == '__main__':
    main()
