import numpy as np


def dlt_estimate(P_array, u_array):
    """
    Solve for unknown position x \in \reals^3
    Observe u_i \in \reals^2 for pixel values
    and projection matrices P_i \reals^{3,4}.

    (u_i; 1) * d_i = P_i (x; 1)

    where d_i is the scale factor for the observation (depth).
    The direct linear transform (DLT) method uses a skew symmetric matrix H
    to remove the d_i from each equation.

    Take H = [0 1; -1 0] so u_i'Hu_i = 0 by skew symmetry. Hence
    (taking the {2,4} relevant section of P_i, that is P[:2, :]
    since we don't need the 1 on lhs):

    0 = (u_i'H)u_i * d_i = (u_i'H) P_i (x; 1)
                         = (u_i'H) (P_i[:2, :3]x + P_i[:2, 3])

    now solve via least squares.

    :param P_array: k, 3, 4
    :param u_array: k, 2
    :return:
    """
    uH = u_array.dot(np.array([[0, -1], [1, 0.]]))  # k, 2
    uHP = np.einsum('li,lik->lk', uH, P_array[:, :2, :3])  # k, 3
    uH_pvec = np.einsum('li,li->l', uH, P_array[:, :2, 3])  # k,
    x = np.linalg.lstsq(uHP, -uH_pvec, rcond=None)[0]
    return x
