from controls import pid
import numpy as np


def build_boat_rotate_move_controller():
    kp_r, ki_r, kd_r = 0.1, 0, 0
    kp_m, ki_m, kd_m = 0.05, 0, 0
    # np.unwrap: https://github.com/numpy/numpy/issues/7494
    state2yaw_err_fcn = lambda x, x_goal: np.unwrap(
        [0, x[4] - np.arctan2(x_goal[1] - x[1], x_goal[0] - x[0])])[1]
    state2pos_err_fcn = lambda x, x_goal: np.linalg.norm(x[:2] - x_goal[:2])
    a_rotate = np.array([1, -1, -1, 1])
    a_translate = np.array([1, 1, 1, 1])
    tau_yaw_err = 0.1
    pid_controller = RotateMovePidController(
        kp_r, ki_r, kd_r, kp_m, ki_m, kd_m,
        state2yaw_err_fcn, state2pos_err_fcn,
        a_rotate, a_translate, tau_yaw_err
    )
    return pid_controller


class RotateMovePidController(object):
    """
    Split motion into rotation followed by translation to reach each goal
    """

    def __init__(self, kp_r, ki_r, kd_r, kp_m, ki_m, kd_m,
                 state2yaw_err_fcn, state2pos_err_fcn,
                 a_rotate, a_translate, tau_yaw_err):
        """
        :param kp_r:
        :param ki_r:
        :param kd_r:
        :param kp_m:
        :param ki_m:
        :param kd_m:
        :param state2yaw_err_fcn:
        :param state2pos_err_fcn:
        :param a_rotate: k, | take 1d output to full u dimension for rotation
        :param a_translate: k, | take 1d output to full u dimension for translation
        :param tau_yaw_err:
        """
        self.rotate_controller = pid.PidController(kp_r, ki_r, kd_r, state2yaw_err_fcn, a_rotate)
        self.pos_controller = pid.PidController(kp_m, ki_m, kd_m, state2pos_err_fcn, a_translate)
        self.state2yaw_err_fcn = state2yaw_err_fcn
        self.state2pos_err_fcn = state2pos_err_fcn
        self.tau_yaw_err = tau_yaw_err
        self.x_goal = ()

    def set_setpoint(self, x_goal):
        self.x_goal = x_goal
        self.rotate_controller.set_setpoint(x_goal)
        self.pos_controller.set_setpoint(x_goal)

    def reset_state(self):
        self.rotate_controller.reset_state()
        self.pos_controller.reset_state()

    def __call__(self, x, t, **kwargs):
        yaw_err = self.state2yaw_err_fcn(x, self.x_goal)
        if np.abs(yaw_err) > self.tau_yaw_err:
            return self.rotate_controller(x, t)
        return self.pos_controller(x, t)
