import numpy as np


def build_nd_interpolator(x, t, interp_builder_fcn, **kwargs):
    """

    :param x: k, n | n-dimensional states over k steps
    :param t: k, | time at each step
    :param interp_builder_fcn: for y, t -> f, build f, where y = f(t) and y is 1d
    :return: spl where x = spl(t) and x is n-dimensional
    """
    spl_1d_list = [interp_builder_fcn(t, x[:, i], **kwargs)
                   for i in range(x.shape[1])]

    def spl(t_):
        return np.array([spl_i(t_) for spl_i in spl_1d_list]).T
    return spl


def order_circular_1d(x, c):
    """
    Process sequence of circular values so that euclidean space
    interpolation will perform the "shortest path" interpolation
    eg: x = [359 1 5 1 359]
    -> y = [359 360+1 360+5 359]
    :param x: n, | values in circular space (eg S^1) to interpolate
    :param c: period of circular space [0, c) (eg 2pi)
    :return:
    """
    y = x % c  # ensure y[0] \in [0, c)
    # y = x.copy()  # preserve first value in sequence
    b = np.array([-c, 0, c])
    for i in range(1, y.size):
        y[i:] = y[i:] % c + c * (y[i-1] // c)
        ind = np.argmin(np.abs(y[i] + b - y[i-1]))
        y[i] += b[ind]
    return y


def _example_order_circular_1d():
    x = np.array([359, 1, 5, 1, 359])
    # [359 361 365 361 359]
    print(order_circular_1d(x, 360))

    x = np.array([359, 1, 179, 182, 1])
    # [359 361 539 542 721]
    print(order_circular_1d(x, 360))

    x = np.array([-1-360, 1, 5, 1, 359])
    # [-361 -359 -355 -359 -361]
    print(order_circular_1d(x, 360))


if __name__ == '__main__':
    _example_order_circular_1d()
