import numpy as np

EPS = np.finfo(float).eps
EPSrt4 = EPS ** (1/4)


def dot_mult_matrix(q):
    """
    Quaternion dot multiplication has an equivalent matrix multiplication:
    q \odot [x; 0] = C(q)x  (eqn 2.98 in MC)

    with C(q) =  ( q_4 I_3 + [q_{1:3} \cross]
                        -q_{1:3}'             )
    :param q: 4, | quaternion
    :return: 4, 3 | C(q) matrix
    """
    C = np.zeros((4, 3))
    C[:3, :3] = cross_matrix(q[:3])
    C[[0, 1, 2], [0, 1, 2]] = q[3]
    C[-1] = -q[:3]
    return C


def quaternion2rotation_matrix(q):
    """
    Rotating a vector has an equivalent matrix multiplication:
    (q \kron {x ; 0} \kron q*)[:3] = A(q)x

    with A(q) = (q_4^2 - ||q_{1:3}||^2) I_3
                - 2 q_4 [q_{1:3} \cross]
                + 2 q_{1:3} q_{1:3}'       (eqn 2.125 in MC)
    :param q: 4, | quaternion
    :return: 3, 3 |
    """
    n = np.linalg.norm(q[:3])
    A = (q[3] ** 2 - n ** 2) * np.eye(3) \
        - 2 * q[3] * cross_matrix(q[:3]) \
        + 2 * np.outer(q[:3], q[:3])
    return A


def exp_map2rotation_matrix(v):
    """
    :param v: 3, | exponential map representation of rotation
    :return: 3, 3
    """
    q = exp_map2quaternion(v)
    A = quaternion2rotation_matrix(q)
    return A


def exp_map2quaternion(v):
    """
    Convert exponential map representation to quaternion,
    where v \in \reals^3 that encodes a unit axis and the angle
    that is rotated around the axis.

    ||v||_2 = \theta = the angle [rad]
    v / ||v||_2 = unit axis

    This is identified by setting \theta = 0 (no rotation) at v = 0,
    but there is numerical error in dividing by ||v||_2 as v -> 0.

    As proposed in (Grassia '99), achieve loss-less conversion by
    using the taylor expansion of the sinc function.
    S. F. Grassia. "Practical parameterization of rotations using the
    exponential map," Journal of Graphics Tools, vol. 3 no. 3, pp.29-48, 1998.

    :param v: 3, | exponential map representation of rotation
    :return: q: 4, | quaternion
    """
    theta = np.linalg.norm(v)
    q = np.zeros((4,))
    if theta > EPSrt4:
        q[:3] = np.sin(theta / 2) / theta
    else:
        q[:3] = 0.5 + theta ** 2 / 48
    q[:3] *= v
    q[3] = np.cos(theta / 2)
    return q


def quaternion2exp_map(q):
    """
    :param q: 4, | quaternion (not necessarily unit norm)
    :return: v: 3, | exponential map representation of rotation
    """
    q_ = 1 * q
    n = np.linalg.norm(q)
    if n > 1:
        # since arc cos range as [-1, 1]
        q_ /= n
    theta = 2 * np.arccos(q_[-1])
    if np.abs(theta) < EPS:
        return np.zeros((3,))
    v = q_[:3] * theta / np.sin(theta / 2)
    return v


def reset_exp_map(v):
    """
    Reset proposed in (Grassia '99)
    :param v: 3, | exponential map representation of rotation
    :return:
    """
    theta = np.linalg.norm(v)
    if theta < np.pi:
        return v
    return v * (1 - 2*np.pi / theta)


def cross_matrix(a):
    """
    a \cross x = [a \cross] x = C(a)x
    where
    C(a) = [a \cross]
    :param a: 3,
    :return: 3, 3 | C(a)
    """
    C = np.array([
        [0, -a[2], a[1]],
        [a[2], 0, -a[0]],
        [-a[1], a[0], 0],
    ])
    return C
