import numpy as np
from rigid_body_models import triangulated_surface as trs

SHUTTLE_VERTICES = np.array([
    [0, 2, 0],
    [.5, 1, 1],  # 1->9
    [1, 0, 1.5],
    [.5, -1, 1],
    [0, -2, 0],
    [9, 0, 1.5],  # 5->12
    [9.5, .5, 1],  # 6->13
    [9.5, -.5, 1],  # 7->14
    [10, 0, 0],
    # bottom points
    [.5, 1, 0],  # 9
    [1, 0, 0],
    [.5, -1, 0],
    [9, 0, 0],
    [9.5, .5, 0],  # 13
    [9.5, -.5, 0],
])
SHUTTLE_TRIANGLES = np.array([
    [0, 1, 6],
    [1, 5, 6],
    [1, 2, 5],
    [3, 7, 5],
    [2, 3, 5],
    [3, 4, 7],
    [5, 7, 6],
    [6, 7, 8],
    # bottom triangles
    [0, 13, 9],
    [9, 13, 12],
    [9, 12, 10],
    [11, 12, 14],
    [10, 12, 11],
    [11, 14, 4],
    [12, 13, 8],
    [12, 8, 14],
    # sides
    [0, 6, 13],
    [4, 14, 7],
    [6, 8, 13],
    [7, 14, 8],
    # back
    [0, 9, 1],
    [1, 9, 10],
    [1, 10, 3],
    [1, 3, 2],
    [10, 11, 3],
    [11, 4, 3],
])


class Shuttle(trs.TriangulatedSurface):
    """
    Body coordinate frame:
        ^ y   (z-up, so right-handed)
        |
       ---
      |    >  --> x
       ---
    """

    r = np.array([
        [0.45, 0.45, 0.4, 0.4],
        [0.8, -0.8, 0.6, -0.6],
        [0.3, 0.3, 0.8, 0.8],
    ])
    f = np.array([
        [1, 1, 1, 1],
        [0, 0, 0, 0],
        [0, 0, 0, 0.],
    ])

    def __init__(self, rho=1e6, vertices=SHUTTLE_VERTICES, triangles=SHUTTLE_TRIANGLES):
        super(Shuttle, self).__init__(rho, vertices, triangles)
        m, center, im = trs.calculate_mass_properties(
            self.vertices, self.triangles, rho)
        self.m = m
        self.im = im
        self.center = 0 * center
        self.vertices -= center
        self.r = (self.r - center[:, np.newaxis])
        # recalculate inertia matrix after translating to center
        _, _, im = trs.calculate_mass_properties(
            self.vertices, self.triangles, rho)
        self.im = im
        self.color = 'white'
