import numpy as np
from rigid_body_models import triangulated_surface as trs
from scipy.spatial import ConvexHull


def generate_convex_asteroid(mu_list, sigma_list, n_samples_per, **kwargs):
    """
    Generate asteroid from the convex hull of samples taken
    from a list of normals.
    :param mu_list: n, 3 | means of multivariate normals
    :param sigma_list: n, 3, 3 | covariance matrix of multivariate normals
    :param n_samples_per: number of samples per normal
    :return:
        vertices:
        triangles:
    """
    samples = []
    for mu, sigma in zip(mu_list, sigma_list):
        L = np.linalg.cholesky(sigma)
        samples_i = (L.dot(np.random.randn(3, n_samples_per))).T + mu
        samples.append(samples_i)
    samples = np.vstack(samples)  # n*n_samples_per, 3
    hull = ConvexHull(samples)
    triangles = hull.simplices
    vertices = samples
    vertices, triangles = trs.remove_unused_vertices(vertices, triangles)
    triangles = trs.make_convex_shape_triangles_consistent(vertices, triangles)
    asteroid = Asteroid(vertices, triangles, **kwargs)
    return asteroid


class Asteroid(trs.TriangulatedSurface):

    def __init__(self, vertices, triangles, rho=2000.):
        super(Asteroid, self).__init__(rho, vertices, triangles)
        m, center, im = trs.calculate_mass_properties(
            self.vertices, self.triangles, rho)
        self.m = m
        self.im = im
        self.center = 0 * center
        self.vertices -= center
        _, _, im = trs.calculate_mass_properties(
            self.vertices, self.triangles, rho)
        self.im = im
        self.color = 'grey'
