"""
Use the Newton-Euler equations and simplifications resulting from:

1) Object motion occurs only in 2d plane
- cross product formulas are simplified
- torques restricted to the z-axis (following right-handed axes, plane is xy)
2) Moment of inertia is a multiple of identity matrix
- angular momentum is zero

Object state defined by:

x_t: 6, | (p_t, v_t, \theta_t, \omega_t)
p_t: 2, | 2d position of center of mass [m]
v_t: 2, | velocity of center of mass [m/s]
\theta_t: yaw [rad]
\omega_t: yaw rate [rad/s]

Motion of rigid body is determined by thrusters that apply force u_i
"""
import numpy as np
from misc import matrix_building as mb


class EmpiricalDragCts(object):
    """
    Simulate drag with

    F_drag = a[0]v_t + a[1]v_t||v_t||_2
    F_angular_drag = b[0]\omega_t + b[1]\omega_t |\omega_t|

    where a, b >= 0 are the supplied drag and angular drag coefficients.
    Thrusters in the plane contribute via
    Bc = ( 0_{2,k}   =  ( 0_{2,k}
           B_t            rot(\theta_t)f/m
           0_{1,k}        0_{1,k}
           c'      )      c'      )
    where c'u = 1/i_m \sum_{i=1}^n (-r_{i,2} ; r{i,1})'f_i u_i
    The interpretation is that Bc applies forces for linear momentum
    in the global frame using current rotation, and torques in the local frame
    since no flips result from rotations in the plane.
    The resulting transition equation is

    x_dot = ( v_t
              -F_drag           +   Bc_t u
              \omega_t
              -F_angular_drag )

          = ( 0 0 1 0 0 0          ( 0_{2,1}
              0 0 0 1 0 0            F_drag
              0 0 0 0 0 0    x  -    0                   +   Bc_t u
              0 0 0 0 0 0            F_angular_drag )
              0 0 0 0 0 1
              0 0 0 0 0 0 )

    This system is affine in the control u since it can be written
    x_dot = f(x)x + g(x)u (here, = [Ax - F(x)] + B(x)u)
    """
    def __init__(self, mounted_thruster_model, drag_c, angular_drag_c):
        self.m = 6
        self.Ac = np.zeros((self.m, self.m))
        self.Ac[:2, 2:4] = np.eye(2)
        self.Ac[4, -1] = 1.
        f = mounted_thruster_model.f
        r = mounted_thruster_model.r
        self.n = f.shape[1]
        self.b_part = 1/mounted_thruster_model.m * f
        self.Bc_t = np.zeros((self.m, self.n))
        self.Bc_t[-1, :] = (-r[1, :] * f[0, :] + r[0, :] * f[1, :]) / \
            mounted_thruster_model.im
        assert (drag_c >= 0).all() and (angular_drag_c >= 0).all()
        self.drag_c = drag_c
        self.angular_drag_c = angular_drag_c

    def step(self, x_t, u_t):
        operators = self.get_process_operators(x_t)
        fx, gx = operators['fx'], operators['gx']
        x_dot = fx + gx.dot(u_t)
        return x_dot

    def get_process_operators(self, x_t):
        B_t = self.Bc_t
        B_t[2:4, :] = mb.rot_2d(x_t[4]).dot(self.b_part)
        A = self.Ac
        f_drag = np.zeros((self.m,))
        norm_v = np.linalg.norm(x_t[2:4])
        f_drag[2:4] = (self.drag_c[0] +
                       self.drag_c[1] * norm_v) * x_t[2:4]
        abs_omega = np.abs(x_t[5])
        f_drag[5] = (self.angular_drag_c[0] +
                     self.angular_drag_c[1] * abs_omega) * x_t[5]
        fx = A.dot(x_t) - f_drag
        gx = B_t
        
        Ca = np.zeros((2, self.m))
        dfxdx_drag = self.drag_c[0] * Ca +\
            self.drag_c[1] * Ca * norm_v +\
            self.drag_c[1] * np.outer(x_t[2:4], x_t[2:4].dot(Ca) / max(norm_v, 1e-8))
        Cb = np.zeros((1, self.m))
        dfxdx_angular_drag = \
            self.angular_drag_c[0] * Cb + \
            self.angular_drag_c[1] * Cb * abs_omega + \
            self.angular_drag_c[1] * x_t[5] * x_t[5] * Cb / max(abs_omega, 1e-8)
        dfxdx_all_drag = np.zeros((self.m, self.m))
        dfxdx_all_drag[2:4, :] = dfxdx_drag
        dfxdx_all_drag[5, :] = dfxdx_angular_drag
        dfxdx = A - dfxdx_all_drag
        return dict(A=A, B=B_t, fx=fx, gx=gx, dfxdx=dfxdx)

    def reset_state(self, x_t):
        return x_t

    def get_mn(self):
        return self.m, self.n
