import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import matplotlib.patches as pa
import misc.matrix_building as mb


def loop_camera_sim(model, xy, theta, u, dt, scene, camera, **kwargs):
    """

    :param model:
    :param xy: k, 2 | positions in xy of (center of mass of) rigid body
    :param theta: k, | orientation of rigid body
    :param u: k, n | control inputs
    :param dt: time between sampled positions
    :param scene:
    :param camera:
    :return:
    """
    fig, ax = plt.subplots(1, 2, figsize=(8, 4))
    k = theta.size
    ax[0].set_xticks(np.mgrid[-50:50:1])
    ax[0].set_yticks(np.mgrid[-50:50:1])
    ax[0].set_xlim([-3, 3])
    ax[0].set_ylim([-3, 3])
    ax[0].set_aspect('equal')
    ax[1].set_aspect('equal')
    ax[0].grid(True)
    ax[0].set_axisbelow(True)

    artists = draw_model(ax[0], model, xy[0], theta[0])
    scene.draw_top_view(ax[0])
    time_text = ax[0].text(
        0.5, 1.05, 't = {:2.2f}s'.format(0), fontsize=14,
        horizontalalignment='center',
        verticalalignment='center', transform=ax[0].transAxes)

    def draw_fcn(i, artists):
        ax[0].set_xlim(xy[i, 0] + [-3, 3])
        ax[0].set_ylim(xy[i, 1] + [-3, 3])
        time_text.set_text('t = {:2.2f}s'.format(i * dt))
        draw_model(ax[0], model, xy[i], theta[i], u[i], artists=artists)
        ax[1].clear()
        body2world_Rt = mb.Rt_planar_xyz1(theta[i], xy[i])
        camera.set_body2world_Rt(body2world_Rt)
        scene.draw_image_view(ax[1], camera)

    ani = animation.FuncAnimation(
        fig, draw_fcn, k, fargs=(artists,), interval=100, repeat_delay=500, blit=False)
    # ani.save('demo.gif', writer=animation.PillowWriter(fps=10))
    plt.show()


def loop_sim(model, xy, theta, u, dt, scene=None, **kwargs):
    """

    :param model:
    :param xy: k, 2 | positions in xy of (center of mass of) rigid body
    :param theta: k, | orientation of rigid body
    :param u: k, n | control inputs
    :param dt: time between sampled positions
    :param scene:
    :return:
    """
    fig, ax = plt.subplots()
    k = theta.size
    ax.set_xticks(np.mgrid[-50:50:1])
    ax.set_yticks(np.mgrid[-50:50:1])
    ax.set_xlim([-3, 3])
    ax.set_ylim([-3, 3])
    ax.set_aspect('equal')
    plt.grid(True)
    ax.set_axisbelow(True)

    artists = draw_model(ax, model, xy[0], theta[0])
    if scene:
        scene.draw_top_view(ax)
    time_text = ax.text(0.5, 1.05, 't = {:2.2f}s'.format(0), fontsize=14,
                        horizontalalignment='center',
                        verticalalignment='center', transform=ax.transAxes)

    def draw_fcn(i, artists):
        ax.set_xlim(xy[i, 0] + [-3, 3])
        ax.set_ylim(xy[i, 1] + [-3, 3])
        time_text.set_text('t = {:2.2f}s'.format(i * dt))
        draw_model(ax, model, xy[i], theta[i], u[i], artists=artists)

    ani = animation.FuncAnimation(
        fig, draw_fcn, k, fargs=(artists,), interval=100, repeat_delay=500, blit=False)
    plt.show()


def draw_model(ax, model, xy, theta, u=(), artists=None):
    """

    :param ax:
    :param model:
    :param xy: 2, | center of model
    :param theta:
    :param u: n,
    :param artists:
    :return:
    """
    if not artists:
        artists = {}

    body_rect_xy = xy - mb.rot_2d(theta).dot(np.array([model.length, model.width])/2)
    body_rect_angle = theta * 180/np.pi
    if 'body_rect' in artists:
        artists['body_rect'].set_xy(body_rect_xy)
        artists['body_rect'].angle = body_rect_angle
    else:
        body_rect = pa.Rectangle(
            body_rect_xy,
            model.length,
            model.width,
            angle=body_rect_angle,
        )
        ax.add_artist(body_rect)
        artists['body_rect'] = body_rect

    if 'body_wedge' in artists:
        artists['body_wedge'].set_center(xy)
        artists['body_wedge'].set_theta1(body_rect_angle + 180 - 20)
        artists['body_wedge'].set_theta2(body_rect_angle + 180 + 20)
    else:
        body_wedge = pa.Wedge(
            xy, .2,
            body_rect_angle + 180 - 20,
            body_rect_angle + 180 + 20,
            zorder=10,
            color='black'
        )
        ax.add_artist(body_wedge)
        artists['body_wedge'] = body_wedge

    if len(u) > 0:
        for i in range(len(u)):
            tag = 'u_{}'.format(i)
            if tag in artists:
                artists[tag].remove()
            artists[tag] = draw_model_thrust(ax, model, xy, theta, u[i], i)
    return artists


def draw_model_thrust(ax, model, xy, theta, u, index):
    """
    Draw a single thrust vector
    :param ax:
    :param model:
    :param xy: 2,
    :param theta:
    :param u: force applied by this thruster
    :param index: index of thruster in model
    :return:
    """
    rot = mb.rot_2d(theta)
    f_world = rot.dot(model.f[:, index])
    xy0 = rot.dot(model.r[:, index]) + xy
    dxy0 = f_world * u
    arrow = pa.Arrow(xy0[0], xy0[1], dxy0[0], dxy0[1],
                     color='green', width=np.log(1 + np.abs(u)))
    return ax.add_artist(arrow)


